/*
 * File:   Worker.hpp
 * Author: John Lång (john.larry.lang@cern.ch)
 *
 * Created on 16 January 2017, 12:37
 */

#ifndef ADAPRO_WORKER_HPP
#define ADAPRO_WORKER_HPP

#include <string>
#include <map>
#include <memory>
#include <functional>
#include "../data/Typedefs.hpp"
#include "./Thread.hpp"

class SupervisorTest;

namespace ADAPRO
{
namespace Data
{
    enum State;
}

namespace Library
{
    template<> inline std::string show(const ADAPRO::Data::State state);
    ADAPRO::Data::config_t make_default_configuration() noexcept;
}

namespace Control
{
#ifndef VERIFY
    class Logger;
#endif
    class Supervisor;

    /**
     * Worker is a Thread that accepts a configuration consisting of key-value
     * pairs.
     */
    class Worker: public Thread
    {
        friend Supervisor;
        friend SupervisorTest;

        /**
         * Number of missed checkpoints, that is how many times Supervisor has
         * observed that this Thread has failed to set the <tt>responsive</tt>
         * to <tt>true</tt> during the one-second time window.
         */
        uint64_t missed_checkpoints;

        /**
         * Number of missed checkpoints, before Supervisor prints a debug
         * message. This treshold is increased every time it is exceeded.
         */
        uint64_t missed_checkpoints_treshold;

    protected:

        /**
         * The configuration of the Worker. <em>It is a reference to a map, so
         * the state of the map object may be changed by other threads during
         * runtime.</em>
         */
        const ADAPRO::Data::config_t& configuration;

    public:

        /**
         * Returns a report similar to the one generated by a Thread, except for
         * an additional entry for missed checkpoint count.
         *
         * @return The Worker's report.
         * @see ADAPRO::Control::Thread::report
         */
        virtual string report() noexcept override
        {
            // Perhaps I could use new lines and indentation...
            return name + ": state = " + ADAPRO::Library::show(get_state()) +
                    ", command = " + ADAPRO::Library::show(get_command()) +
                    ", missed checkpoints = " + to_string(missed_checkpoints);
        }

        /**
         * The deleted default constructor for Worker.
         */
        Worker() = delete;

        /**
         * Constructs a Worker.
         *
         * @param logger                The Logger instance of the ADAPRO
         * Session.
         * @param name                  Name of the Worker Thread.
         * @param configuration         The ADAPRO configuration.
         * @param preferred_core        Preferred core.
         * @param transition_callback   Transition callback.
         *
         * @see ADAPRO::Control::Thread
         * @see ADAPRO::Control::Supervisor
         */
        Worker
        (
#ifndef VERIFY
                Logger& logger,
#endif
                std::string&& name,
                const ADAPRO::Data::config_t& configuration,
                const int preferred_core = -1,
                const ADAPRO::Data::trans_cb_t transition_callback =
                        [](const ADAPRO::Data::State s){}
        )
        noexcept:
                ADAPRO::Control::Thread
                (
#ifndef VERIFY
                        logger,
#endif
                        std::move(name),
                        transition_callback,
                        preferred_core
                ),
                missed_checkpoints(0),
                missed_checkpoints_treshold(10),
                configuration(configuration) {}

        virtual ~Worker() noexcept {}
    };
}
}

#endif /* ADAPRO_WORKER_HPP */

