/*
 * File:   ConfigurationUtils.hpp
 * Author: John Lång (john.larry.lang@cern.ch)
 *
 * Created on 17 January 2017, 15:20
 */

#ifndef ADAPRO_CONFIGURATION_UTILS_HPP
#define ADAPRO_CONFIGURATION_UTILS_HPP

#include <string>
#include <list>
#include <memory>
#include <functional>
#include <map>
#include <utility>
#include "../data/Parameters.hpp"
#include "../data/Typedefs.hpp"

namespace ADAPRO
{
namespace Data
{
    struct Context;
}
#ifndef VERIFY
namespace Control
{
    class Logger;
}
#endif

namespace Library
{
    /**
     * <p>Makes sure that the mandatory entries exist in the configuration. If
     * they don't, an exception will be thrown. The keys for the currently
     * mandatory configuration entries are:</p>
     * <ul>
     *     <li><tt>DAEMON_ENABLED</tt>. If the value of this entry is
     *         <tt>"TRUE"</tt>, then also <tt>DAEMON_LOGFILE</tt> is mandatory.
     *     </li>
     *     <li><tt>DIM_SERVER_ENABLED</tt>.If the value of this entry is
     *         <tt>"TRUE"</tt>, then also the following enries are mandatory:
     *         <ul>
     *             <li><tt>DIM_DNS_NODE</tt>;</li>
     *             <li><tt>DIM_DNS_PORT</tt>;</li>
     *             <li><tt>DIM_CALLBACK_CORE</tt> (on Linux systems);</li>
     *             <li><tt>DIM_CONTROLLER_CORE</tt> (on Linux systems); and</li>
     *             <li><tt>DIM_SERVER_NAME</tt>.</li>
     *         </ul>
     *     </li>
     *     <li><tt>LOGGING_MASK</tt></li>
     *     <li><tt>COLOURS_ENABLED</tt></li>
     *     <li><tt>SERIALIZE_COMMANDS</tt></li>
     * </ul>
     * <p>Depending on operating system support, the entries for
     * <tt>LOCK_MEMORY</tt>, <tt>MAIN_CORE</tt>, <tt>NICE</tt>, and
     * <tt>SUPERVISOR_CORE</tt> may be also mandatory.</p>
     *
     * @since ADAPRO 5.0.0
     * @param configuration The configuration to check.
     * @throws std::invalid_argument if any of the mandatory configuration
     * entries were missing.
     * @see ADAPRO::Data::COLOURS_ENABLED
     * @see ADAPRO::Data::DAEMON_ENABLED
     * @see ADAPRO::Data::DAEMON_LOGFILE
     * @see ADAPRO::Data::DIM_CALLBACK_CORE
     * @see ADAPRO::Data::DIM_CONTROLLER_CORE
     * @see ADAPRO::Data::DIM_DNS_NODE
     * @see ADAPRO::Data::DIM_DNS_PORT
     * @see ADAPRO::Data::DIM_SERVER_ENABLED
     * @see ADAPRO::Data::DIM_SERVER_NAME
     * @see ADAPRO::Data::LOCK_MEMORY
     * @see ADAPRO::Data::LOGGING_MASK
     * @see ADAPRO::Data::MAIN_CORE
     * @see ADAPRO::Data::NICE
     * @see ADAPRO::Data::SERIALIZE_COMMANDS
     * @see ADAPRO::Data::SUPERVISOR_CORE
     */
    void check_mandatory_entries(const config_t& configuration);

    /**
     * <p>Imports the map of key-value pairs from the given file, if possible.
     * If not, an exception is thrown to indicate the error. Entries with
     * invalid values in the configuration file will not be updated in the
     * configuration map. <em>By the time a potential exception generated by an
     * invalid entry is thrown, all entries with valid values in the file, will
     * be fully updated in the map, and are guaranteed to satisfy the relation
     * and.</em> This procedure doesn't rollback to the initial state of the
     * given configuration.</p>
     * <p>The expected format of the key-value file is the following:</p>
     * <pre>
     * &lt;key-value-file&gt;   ::= &lt;line&gt; "\\n" &lt;key-value-file&gt;
     * &lt;line&gt;             ::= &lt;key&gt; &lt;whitespace&gt; &lt;key&gt; | ""
     * </pre>
     * <p>
     * where
     * </p>
     * <ol>
     *     <li><tt>&lt;key&gt;</tt> needs to satisfy the regular expression
     *         <tt>[a-zA-Z0-9_]+</tt>;</li>
     *     <li><tt>&lt;whitespace&gt;</tt> must be a string with length one of
     *         more, consisting of ASCII characters <tt>0x20</tt> (space) and
     *         <tt>0x09</tt> (horizontal tab); and</li>
     *     <li><tt>&lt;value&gt;</tt> must not contain whitespace characters
     *         (including newline, carriage return, etc.).</li>
     * </ol>
     * <p>
     * The character '#' and its successors in a single line will be ignored.
     * Each key can be only associated with one value.
     * </p>
     *
     * @param initial_configuration The map into which the key-value pairs
     * loaded from the configuratoin file will be stored. Any existing values
     * will be overwritten in case of collision.
     * @param filenames             A <em>non-empty</em> vector containing
     * filename candidates for importing the map. This function will proceed
     * with the first accessible filename on the list or throw an exception if
     * none of them can be accessed. <em>The well-formedness of filenames is not
     * checked. This means that illegal filenames are treated as if they were
     * legal but the files they refer to were inaccessible.</em>
     * @param allow_new_keys        If <tt>true</tt>, then the file is allowed
     * to contain keys that were not in the map already. Otherwise, an exception
     * is thrown if a previously unknown key is encountered. <em>This function
     * treats keys associated with empty string in the map, in the same way as
     * non-existing keys.</em>
     * @param relation              <p>Relation that each key-value pair needs
     * to satisfy. Only the key-value pairs satisfying the relation will be
     * added to the map.</p>
     * <p>The implementation of the relation is expected to be a function
     * (e.g. a lambda expression) returning <tt>true</tt> for every pair of
     * <tt>std::string</tt> satisfying some conditions and <tt>false</tt> for
     * others. The default relation ensures the well-formedness of ADAPRO
     * configuration entries. For other key-value pairs, the default relation
     * acts like the Cartesian product (accepting them as they are).</p><p>Note
     * that a relation (in mathematical sense) is not capable of enforcing
     * restrictions on values that depend on multiple keys.</p>
     * @return The name of the configuration file accessed (if any), from the
     * list of filenames given as the argument <tt>filenames</tt>. <em>NB: This
     * reference is valid if and only if the list still exists after this
     * routine, i.e. if it wasn't a temporary object.</em>
     * @throws std::invalid_argument    The list of configuration file locations
     * was empty.
     * @throws std::domain_error        This exception is thrown in the
     * following situations:
     * <ol>
     *     <li>If the file fails to satisfy the conditions stated above;</li>
     *     <li>if the file contained unknown keys that were forbidden; or</li>
     *     <li>if multiple definitions for the same key were found in the file.</li>
     * </ol>
     * @throws std::ios_base::failure   If none of the key-value files could be
     * accessed.
     * @see ADAPRO::Library::default_relation
     */
    const std::string& import_configuration
    (
            ADAPRO::Data::config_t& initial_configuration,
            const std::list<std::string>& filenames,
            const bool allow_new_keys = false,
            const std::function<bool(const std::string&, const std::string&)>&
                    relation = ADAPRO::Library::default_relation
    );

    /**
     * Queries (but does not verify) the configuration for the parameter.
     *
     * @param configuration The configuration to be queried.
     * @param parameter     The parameter, whose value will be returned.
     * @return The value of the queried configuration parameter.
     * @throws std::out_of_range if the given parameter was not found in the
     * configuration
     * @throws std::domain_error if the given parameter was invalid (i.e. a
     * numerical value that did not represent a valid <tt>Parameter</tt> value).
     * @since ADAPRO 5.0.0
     * @see ADAPRO::Data::parameter_t
     * @see ADAPRO::Library::verify
     */
    inline const std::string& query
    (
            const ADAPRO::Data::config_t& configuration,
            const ADAPRO::Data::Parameter parameter
    )
    {
        return configuration.at(show(parameter));
    }

    /**
     * Writes the given map to a file. The output will be written in a format
     * that is specified in the documentation entry of <tt>import_map</tt>.
     * <em>Verification of the entries is performed before writing them to the
     * output file. In case of an exception due to relation failure, the output
     * file will remain unmodified.</em>
     *
     * @param key_value_map The map to be exported.
     * @param filename      Name of the output file.
     * @param relation              <p>Relation that each key-value pair needs
     * to satisfy. Only the key-value pairs satisfying the relation will be
     * added to the map.</p>
     * <p>The implementation of the relation is expected to be a function
     * (e.g. a lambda expression) returning <tt>true</tt> for every pair of
     * <tt>std::string</tt> satisfying some conditions and <tt>false</tt> for
     * others. The default relation ensures the well-formedness of ADAPRO
     * configuration entries. For other key-value pairs, the default relation
     * acts like the Cartesian product (accepting them as they are).</p><p>Note
     * that a relation (in mathematical sense) is not capable of enforcing
     * restrictions on values that depend on multiple keys.</p>
     * @param tabulate      If <tt>true</tt>, then the values will be tabulated
     * with spaces to begin at index <em>n</em>+1 at each row, where <em>n</em>
     * is the length of the longest key. For example, tabulating a map
     * <tt>{{"foo", "bar"}, {"very_long", "23"}, {"x", "4"}}</tt> would yield a
     * a file with the following contents:
     * <pre>
     * foo       bar
     * very_long 23
     * x         4
     * </pre>
     * @throws std::invalid_argument If the map contained key-value pairs that
     * would produce output incompatible with <tt>import_map</tt>.
     * @see ADAPRO::Library::import_map
     * @see ADAPRO::Library::default_relation
     */
    void export_configuration
    (
            const ADAPRO::Data::config_t& key_value_map,
            const std::string& filename,
            const bool tabulate = true,
            const std::function<bool(const std::string&, const std::string&)>&
                    relation = ADAPRO::Library::default_relation
    );

    /**
     * A version of <tt>export_map</tt> that prints the map into
     * <tt>stdout</tt>. The output will be written in a format slightly
     * different than what is specified in the documentation entry of
     * <tt>import_map</tt>. The output will be columnated with an automatically
     * computed offset.
     *
     * @param logger        The Logger instance for output.
     * @param title         A heading that is printed before the values.
     * @param key_value_map The map to be printed.
     * @see ADAPRO::Library::import_map
     * @see ADAPRO::Library::export_map
     */
    void print_configuration
    (
#ifndef VERIFY
            ADAPRO::Control::Logger& logger,
#endif
            const std::string& title,
            const ADAPRO::Data::config_t& key_value_map
    )
    noexcept;
}
}

#endif /* ADAPRO_CONFIGURATION_UTILS_HPP */

